//
//  TwilioVoice.h
//  TwilioVoice
//
//  Copyright © 2018 Twilio, Inc. All rights reserved.
//

#import <Foundation/Foundation.h>

#import "TVOAcceptOptions.h"
#import "TVOAudioCodec.h"
#import "TVOAudioDevice.h"
#import "TVOAudioFormat.h"
#import "TVOBaseTrackStats.h"
#import "TVOCall.h"
#import "TVOCallDelegate.h"
#import "TVOCallInvite.h"
#import "TVOCallOptions.h"
#import "TVOCancelledCallInvite.h"
#import "TVOConnectOptions.h"
#import "TVODefaultAudioDevice.h"
#import "TVOError.h"
#import "TVOIceCandidatePairStats.h"
#import "TVOIceCandidateStats.h"
#import "TVOIceOptions.h"
#import "TVOLocalAudioTrackStats.h"
#import "TVOLocalTrackStats.h"
#import "TVONotificationDelegate.h"
#import "TVOOpusCodec.h"
#import "TVOPcmuCodec.h"
#import "TVORemoteAudioTrackStats.h"
#import "TVORemoteTrackStats.h"
#import "TVOStatsReport.h"


/**
 * An enumeration indicating log levels that can be used with the SDK.
 */
typedef NS_ENUM(NSUInteger, TVOLogLevel) {
    TVOLogLevelOff = 0,     ///< Disables all SDK logging.
    TVOLogLevelFatal,       ///< Show very severe error events that will presumably lead the application to abort.
    TVOLogLevelError,       ///< Show errors only.
    TVOLogLevelWarning,     ///< Show warnings as well as all **Error** log messages.
    TVOLogLevelInfo,        ///< Show informational messages as well as all **Warn** log messages.
    TVOLogLevelDebug,       ///< Show debugging messages as well as all **Info** log messages.
    TVOLogLevelTrace,       ///< Show low-level debugging messages as well as all **Debug** log messages.
    TVOLogLevelAll          ///< Show all logging.
};

/**
 * `TwilioVoice` logging modules.
 */
typedef NS_ENUM(NSUInteger, TVOLogModule) {
    TVOLogModuleCore = 0,   ///< Voice Core SDK.
    TVOLogModulePlatform,   ///< Voice iOS SDK.
    TVOLogModuleSignaling,  ///< Signaling module.
    TVOLogModuleWebRTC      ///< WebRTC module.
};

/**
 * `TwilioVoice` is the entry point to the Twilio Voice SDK. You can register for VoIP push notifications, make outgoing
 * Calls, receive Call invites and set audio device to use in a Call using this class.
 */
@interface TwilioVoice : NSObject

/**
 * @brief The current logging level used by the SDK.
 *
 * @discussion The default logging level is `TVOLogLevelError`. `TwilioVoice` and its components use NSLog internally.
 *
 * @see TVOLogLevel
 */
@property (nonatomic, assign, class) TVOLogLevel logLevel;

/**
 *  @brief The `TVOAudioDevice` used to record and playback audio in a Call.
 *
 *  @discussion If you wish to provide your own `TVOAudioDevice` then you must set it before performing any other
 *  actions with the SDK like connecting to a Call. It is also possible to change the device when you are no longer
 *  connected to any Calls and have destroyed all other SDK objects.
 *
 *  @see [TVOAudioDevice](TVOAudioDevice.h)
 */
@property (class, nonatomic, strong, nonnull) id<TVOAudioDevice> audioDevice;

/**
 * @brief Returns the version of the SDK.
 *
 * @return The version of the SDK.
 */
+ (nonnull NSString *)sdkVersion;

/**
 * @brief Sets the logging level for an individual module.
 *
 * @param logLevel The `<TVOLogLevel>` level to be used by the module.
 * @param module The `<TVOLogModule>` for which the logging level is to be set.
 *
 * @see TVOLogLevel
 * @see TVOLogModule
 */
+ (void)setLogLevel:(TVOLogLevel)logLevel module:(TVOLogModule)module;

/**
 *  @brief Retrieve the log level for a specific module in the TwilioVoice SDK.
 *
 *  @param module The `TVOLogModule` for which the log level needs to be set.
 *
 *  @return The current log level for the specified module.
 */
+ (TVOLogLevel)logLevelForModule:(TVOLogModule)module;

/**
 * @name Managing VoIP Push Notifications
 */

/**
   @brief Registers for Twilio VoIP push notifications.
 
   @discussion Registering for push notifications is required to receive incoming call notification messages through Twilio's
   infrastructure. Once successfully registered, the registered binding has a time-to-live(TTL) of 1 year. If the registered binding
   is inactive for 1 year it is deleted and push notifications to the registered identity will not succeed. However, whenever
   the registered binding is used to reach the registered identity the TTL is reset to 1 year. A successful registration will
   ensure that push notifications will arrive via the APNs for the lifetime of the registration device token provided by
   the APNs instance.
 
   @discussion The identity provided in the Access Token may only contain alpha-numeric and underscore
   characters. Other characters, including spaces, will result in undefined behavior.
 
   If the registration is successful the completion handler will contain a `Null NSError`. If the registration
   fails, the completion handler will have a `nonnull NSError` with `UserInfo` string describing the reason for failure.
 
   <table border="1" summary="Registration Errors.">
   <tr>
   <td>Registration Error</td><td>Error Code</td><td>Description</td>
   </tr>
   <tr>
   <td> TVOErrorAccessTokenInvalidError </td><td>20101</td><td>Twilio was unable to validate your Access Token</td>
   </tr>
   <tr>
   <td> TVOErrorAccessTokenHeaderInvalidError </td><td>20102</td><td>Invalid Access Token header</td>
   </tr>
   <tr>
   <td> TVOErrorAccessTokenIssuerInvalidError </td><td>20103</td><td>Invalid Access Token issuer or subject</td>
   </tr>
   <tr>
   <td> TVOErrorAccessTokenExpiredError </td><td>20104</td><td>Access Token has expired or expiration date is invalid</td>
   </tr>
   <tr>
   <td> TVOErrorAccessTokenNotYetValidError </td><td>20105</td><td>Access Token not yet valid</td>
   </tr>
   <tr>
   <td> TVOErrorAccessTokenGrantsInvalidError </td><td>20106</td><td>Invalid Access Token grants</td>
   </tr>
   <tr>
   <td> TVOErrorExpirationTimeExceedsMaxTimeAllowedError </td><td>20157</td><td>Expiration Time in the Access Token Exceeds Maximum Time Allowed</td>
   </tr>
   <tr>
   <td> TVOErrorAccessForbiddenError </td><td>20403</td><td>Forbidden. The account lacks permission to access the Twilio API</td>
   </tr>
   <tr>
   <td> TVOErrorTokenAuthenticationRejected </td><td>51007</td><td>Token authentication is rejected by authentication service</td>
   </tr>
   <tr>
   <td> TVOErrorRegistrationError </td><td>31301</td><td>Registration failed. Look at [error localizedFailureReason] for details</td>
   </tr>
   </table>
 
   **Insights**
 
   <table border="1" summary="Insights events.">
   <tr>
   <td>Group Name</td><td>Event Name</td><td>Description</td><td>Since version</td>
   </tr>
   <tr>
   <td>registration</td><td>registration</td><td>Registration is successful</td><td>3.0.0-beta3</td>
   </tr>
   <tr>
   <td> registration </td> <td>registration-error</td><td>Registration failed</td><td>3.0.0-beta3</td>
   </tr>
   </table>
 
   ** An example of using the *registration* API **
 
   ```
    [TwilioVoice registerWithAccessToken:accessToken
        deviceToken:deviceToken
        completion:^(NSError *error) {
            if(error != nil) {
                NSLog("Failed to register for incoming push: %@\n\t  - reason: %@", [error localizedDescription], [error localizedFailureReason]);
            } else {
                NSLog("Registration successful");
            }
    }];
   ```
 
   @param accessToken Twilio Access Token.
   @param deviceToken The push registry token for Apple VoIP Service.
   @param completion Callback block to receive the result of the registration.
 */
+ (void)registerWithAccessToken:(nonnull NSString *)accessToken
                    deviceToken:(nonnull NSString *)deviceToken
                     completion:(nullable void(^)(NSError * __nullable error))completion;

/**
   @brief Unregisters from Twilio VoIP push notifications.
 
   @discussion Call this method when you no longer want to receive push notifications for incoming Calls.
 
   @discussion The identity provided in the Access Token may only contain alpha-numeric and underscore
   characters. Other characters, including spaces, will result in undefined behavior.
 
   If the unregistration is successful the completion handler will contain a `Null NSError`. If the unregistration
   fails, the completion handler will have a `nonnull NSError` with `UserInfo` string describing the reason for failure.
 
   <table border="1" summary="Registration Errors.">
   <tr>
   <td>Registration Error</td><td>Error Code</td><td>Description</td>
   </tr>
   <tr>
   <td> TVOErrorAccessTokenInvalidError </td><td>20101</td><td>Twilio was unable to validate your Access Token</td>
   </tr>
   <tr>
   <td> TVOErrorAccessTokenHeaderInvalidError </td><td>20102</td><td>Invalid Access Token header</td>
   </tr>
   <tr>
   <td> TVOErrorAccessTokenIssuerInvalidError </td><td>20103</td><td>Invalid Access Token issuer or subject</td>
   </tr>
   <tr>
   <td> TVOErrorAccessTokenExpiredError </td><td>20104</td><td>Access Token has expired or expiration date is invalid</td>
   </tr>
   <tr>
   <td> TVOErrorAccessTokenNotYetValidError </td><td>20105</td><td>Access Token not yet valid</td>
   </tr>
   <tr>
   <td> TVOErrorAccessTokenGrantsInvalidError </td><td>20106</td><td>Invalid Access Token grants</td>
   </tr>
   <tr>
   <td> TVOErrorExpirationTimeExceedsMaxTimeAllowedError </td><td>20157</td><td>Expiration Time in the Access Token Exceeds Maximum Time Allowed</td>
   </tr>
   <tr>
   <td> TVOErrorAccessForbiddenError </td><td>20403</td><td>Forbidden. The account lacks permission to access the Twilio API</td>
   </tr>
   <tr>
   <td> TVOErrorTokenAuthenticationRejected </td><td>51007</td><td>Token authentication is rejected by authentication service</td>
   </tr>
   <tr>
   <td> TVOErrorRegistrationError </td><td>31301</td><td>Registration failed. Look at [error localizedFailureReason] for details</td>
   </tr>
   </table>
 
   **Insights**
 
   <table border="1" summary="Insights events.">
   <tr>
   <td>Group Name</td><td>Event Name</td><td>Description</td><td>Since version</td>
   </tr>
   <tr>
   <td>registration</td><td>unregistration</td><td>Unregistration is successful</td><td>3.0.0-beta3</td>
   </tr>
   <tr>
   <td>registration</td><td>unregistration-error</td><td>Unregistration failed</td><td>3.0.0-beta3</td>
   </tr>
   </table>
 
   ** An example of using the *unregistration* API **
 
   ```
    [TwilioVoice unregisterWithAccessToken:accessToken
        deviceToken:deviceToken
        completion:^(NSError *error) {
            if(error != nil) {
                NSLog("Failed to unregister for incoming push: %@\n\t  - reason: %@", [error localizedDescription], [error localizedFailureReason]);
            } else {
                NSLog("Unregistration successful");
            }
    }];
   ```
   @param accessToken Twilio Access Token.
   @param deviceToken The push registry token for Apple VoIP Service.
   @param completion Callback block to receive the result of the unregistration.
 */
+ (void)unregisterWithAccessToken:(nonnull NSString *)accessToken
                      deviceToken:(nonnull NSString *)deviceToken
                       completion:(nullable void(^)(NSError * __nullable error))completion;

/**
   @brief Processes an incoming VoIP push notification payload.
 
   @discussion This method will synchronously process call notification payload and call the provided delegate
   on the same dispatch queue.
 
   Twilio sends two types of notifications via Apple VoIP Service, `call` and `cancel`.
   The notification type is encoded in the dictionary with the key `twi_message_type` and the values
   `twilio.voice.call` and `twilio.voice.cancel`.
 
   A `call` notification is sent when someone wants to reach the registered `identity`.
   Passing a `call` notification into this method will result in a `callInviteReceived:` callback with a `<TVOCallInvite>`
   object and return `YES`.
 
   A `cancel` notification is sent to an `identity` for the following reasons:
 
   - call made to this `identity` is prematurely `disconnected` by the caller
   - call is `rejected`
   - call is `ignored`
   - after a call is `accepted`. When a call is accepted, a `cancel` notification is sent to all the devices that have
     been registered with the `identity` accepting the call, even to the device that accepted the call. It is recommended
     that in this case the developer checks whether the call sid of `<TVOCancelledCallInvite>` matches the accepted call,
     and if so disregard it.

   Passing a `cancel` notification into this method will result in a `cancelledCallInviteReceived:` callback with
   a `<TVOCancelledCallInvite>` object and return `YES`.
 
   **Insights**
 
   <table border="1" summary="Insights events.">
   <tr>
   <td>Group Name</td><td>Event Name</td><td>Description</td><td>Since version</td>
   </tr>
   <tr>
   <td>connection</td><td>incoming</td><td>Incoming call notification received</td><td>3.0.0-preview1</td>
   </tr>
   </table>
 
   ** An example of using the *handleNotification* API **
 
   ```
    @interface CallViewController () <PKPushRegistryDelegate, TVONotificationDelegate, TVOCallDelegate, ... >
 
    // property declarations
 
    @end
 
    @implementation CallViewController
 
    // ViewController setup and other code ...
 
    #pragma mark - PKPushRegistryDelegate
 
    - (void)pushRegistry:(PKPushRegistry *)registry
        didReceiveIncomingPushWithPayload:(PKPushPayload *)payload
        forType:(NSString *)type {
            NSLog(@"Received incoming push: %@", payload.dictionaryPayload);
 
            if ([type isEqualToString:PKPushTypeVoIP]) {
                if (![TwilioVoice handleNotification:payload.dictionaryPayload delegate:self]) {
                    NSLog(@"The push notification was not a Twilio Voice push notification");
            }
        }
    }
 
    #pragma mark - TVONotificationDelegate methods
 
    - (void)callInviteReceived:(TVOCallInvite *)callInvite {
        // handle call invite
    }
 
    - (void)cancelledCallInviteReceived:(TVOCancelledCallInvite *)cancelledCallInvite {
        // handle cancelled call invite
    }
 
    @end
   ```
   @param payload Push notification payload.
   @param delegate A `<TVONotificationDelegate>` to receive incoming push notification callbacks.
 
   @return A BOOL value that indicates whether the payload is a valid notification sent by Twilio.
 
   @see TVONotificationDelegate
 */
+ (BOOL)handleNotification:(nonnull NSDictionary *)payload
                  delegate:(nonnull id<TVONotificationDelegate>)delegate;


/**
   @brief Make an outgoing Call.
 
   @discussion This method is guaranteed to return a `<TVOCall>` object. It's possible for the returned Call to either
   succeed or fail to connect.
 
   @discussion The identity provided in the Access Token may only contain alpha-numeric and underscore
   characters. Other characters, including spaces, will result in undefined behavior.
 
   The `<TVOCallDelegate>` will receive the Call state update callbacks. The callbacks are listed here.
 
   <table border="1" summary="TVOCallDelegate events.">
   <tr>
   <td>Callback Name</td><td>Description</td><td>Since version</td>
   </tr>
   <tr>
   <td> call:didFailToConnectWithError </td><td>The call failed to connect. An `NSError` object provides details of the root cause.</td><td>3.0.0-preview1</td>
   </tr>
   <tr>
   <td> callDidStartRinging </td><td>Emitted once before the `callDidConnect` callback when the callee is being alerted of a Call.</td><td>3.0.0-preview2</td>
   </tr>
   <tr>
   <td> callDidConnect </td><td>The Call has connected.</td><td>3.0.0-preview1</td>
   </tr>
   <tr>
   <td> call:didDisconnectWithError </td><td>The Call was disconnected. If the Call ends due to an error the `NSError` is non-null. If the call ends normally `NSError` is nil.</td><td>3.0.0-preview1</td>
   </tr>
   </table>
 
   If connect fails, `call:didFailToConnectWithError` or `call:didDisconnectWithError` callback is raised with an `NSError` object. `[error localizedDescription]`,
   `[error localizedFailureReason]` and `[error code]` provide details of the failure.
 
   If any authentication error occurs then the SDK will receive one of the following errors.
 
   <table border="1" summary="Call Connect errors">
   <tr>
   <td>Authentication Errors</td><td>Error Code</td><td>Description</td>
   </tr>
   <tr>
   <td>TVOErrorAccessTokenInvalidError</td><td>20101</td><td>Twilio was unable to validate your Access Token</td>
   </tr>
   <tr>
   <td>TVOErrorAccessTokenHeaderInvalidError</td><td>20102</td><td>Invalid Access Token header</td>
   </tr>
   <tr>
   <td>TVOErrorAccessTokenIssuerInvalidError</td><td>20103</td><td>Invalid Access Token issuer or subject</td>
   </tr>
   <tr>
   <td>TVOErrorAccessTokenExpiredError</td><td>20104</td><td>Access Token has expired or expiration date is invalid</td>
   </tr>
   <tr>
   <td>TVOErrorAccessTokenNotYetValidError</td><td>20105</td><td>Access Token not yet valid</td>
   </tr>
   <tr>
   <td>TVOErrorAccessTokenGrantsInvalidError</td><td>20106</td><td>Invalid Access Token grants</td>
   </tr>
   <tr>
   <td>TVOErrorAuthFailureCodeError</td><td>20151</td><td>Twilio failed to authenticate the client</td>
   </tr>
   <tr>
   <td>TVOErrorExpirationTimeExceedsMaxTimeAllowedError</td><td>20157</td><td>Expiration Time in the Access Token Exceeds Maximum Time Allowed</td>
   </tr>
   <tr>
   <td>TVOErrorAccessForbiddenError</td><td>20403</td><td>Forbidden. The account lacks permission to access the Twilio API</td>
   </tr>
   <tr>
   <td>TVOErrorApplicationNotFoundError</td><td>21218</td><td>Invalid ApplicationSid</td>
   </tr>
   </table>
 
   If any connection fails because of any other error then the SDK will receive one of the following errors.
 
   <table border="1" summary="Call Connect errors">
   <tr>
   <td>Connection Errors</td><td>Error Code</td><td>Description</td>
   </tr>
   <tr>
   <td>TVOErrorConnectionTimeoutError</td><td>31003</td><td>Connection timeout</td>
   </tr>
   <tr>
   <td>TVOErrorConnectionError</td><td>31005</td><td>Connection error</td>
   </tr>
   <tr>
   <td>TVOErrorCallCancelledError</td><td>31008</td><td>Call Cancelled</td>
   </tr>
   <tr>
   <td>TVOErrorMalformedRequestError</td><td>31100</td><td>Malformed request</td>
   </tr>
   <tr>
   <td>TVOErrorAuthorizationError</td><td>31201</td><td>Authorization error</td>
   </tr>
   <tr>
   <td>TVOErrorMicrophoneAccessDenialError</td><td>31208</td><td>User denied access to microphone</td>
   </tr>
   <tr>
   <td>TVOErrorSignalingConnectionError</td><td>53000</td><td>Signaling connection error</td>
   </tr>
   <tr>
   <td>TVOErrorSignalingConnectionDisconnectedError</td><td>53001</td><td>Signaling connection disconnected</td>
   </tr>
   <tr>
   <td>TVOErrorSignalingConnectionTimeoutError</td><td>53002</td><td>Signaling connection timed out</td>
   </tr>
   <tr>
   <td>TVOErrorSignalingIncomingMessageInvalidError</td><td>53003</td><td>Client received an invalid signaling message</td>
   </tr>
   <tr>
   <td>TVOErrorSignalingOutgoingMessageInvalidError</td><td>53004</td><td>Client sent an invalid signaling message</td>
   </tr>
   <tr>
   <td>TVOErrorMediaClientLocalDescFailedError</td><td>53400</td><td>Client is unable to create or apply a local media description</td>
   </tr>
   <tr>
   <td>TVOErrorMediaServerLocalDescFailedError</td><td>53401</td><td>Server is unable to create or apply a local media description</td>
   </tr>
   <tr>
   <td>TVOErrorMediaClientRemoteDescFailedError</td><td>53402</td><td>Client is unable to apply a remote media description</td>
   </tr>
   <tr>
   <td>TVOErrorMediaServerRemoteDescFailedError</td><td>53403</td><td>Server is unable to apply a remote media description</td>
   </tr>
   <tr>
   <td>TVOErrorMediaNoSupportedCodecError</td><td>53404</td><td>No supported codec</td>
   </tr>
   <tr>
   <td>TVOErrorMediaConnectionError</td><td>53405</td><td>Media connection failed</td>
   </tr>
   <tr>
   <td>TVOErrorConfigurationAcquireFailedError</td><td>53500</td><td>Unable to acquire configuration</td>
   </tr>
   <tr>
   <td>TVOErrorConfigurationAcquireTurnFailedError</td><td>53501</td><td>Unable to acquire TURN credentials</td>
   </tr>
   </table>
 
   **Insights**
 
   <table border="1" summary="Insights events.">
   <tr>
   <td>Group Name</td><td>Event Name</td><td>Description</td><td>Since version</td>
   </tr>
   <tr>
   <td>connection</td><td>outgoing</td><td>Outgoing call is made</td><td>3.0.0-beta2</td>
   </tr>
   </table>
 
   If connection fails then an error event will be published.
 
   <table border="1" summary="Insights events.">
   <tr>
   <td>Group Name</td><td>Event Name</td><td>Description</td><td>Since version</td>
   </tr>
   <tr>
   <td>connection</td><td>error</td><td>error description</td><td>3.0.0-beta2</td>
   </tr>
   </table>
 
   ** An example of using the *connectWithAccessToken:delegate:* API **
 
   ```
    @interface CallViewController () <TVOCallDelegate, ... >
 
    // property declarations
 
    @end
 
    @implementation CallViewController
 
    // ViewController setup and other code ...
 
    - (void)makeCall:(NSString*)accessToken
        self.call = [TwilioVoice connectWithAccessToken:accessToken delegate:self];
    }
 
    #pragma mark - TVOCallDelegate
 
    - (void)callDidStartRinging:(TVOCall *)call {
        NSLog(@"Call is ringing at called party %@", call.to);
    }
 
    - (void)call:(TVOCall *)call didFailToConnectWithError:(NSError *)error {
        NSLog("Failed to Connect the Call: %@\n\t  - reason: %@", [error localizedDescription], [error localizedFailureReason]);
    }
 
    - (void)call:(TVOCall *)call didDisconnectWithError:(NSError *)error {
        if (error) {
            NSLog(@"Call disconnected with error: %@", error);
        } else {
            NSLog(@"Call disconnected");
        }
    }
 
    - (void)callDidConnect:(TVOCall *)call {
        NSLog(@"Call connected.");
    }
 
    @end
   ```
 
   @param accessToken The access token.
   @param delegate A `<TVOCallDelegate>` to receive Call state updates.
 
   @return A `<TVOCall>` object.
 
   @see TVOCall
   @see TVOCallDelegate
 */
+ (nonnull TVOCall *)connectWithAccessToken:(nonnull NSString *)accessToken
                                   delegate:(nonnull id<TVOCallDelegate>)delegate;

/**
   @brief Make an outgoing Call.
 
   @discussion This method is guaranteed to return a `<TVOCall>` object. It's possible for the returned Call to either
   succeed or fail to connect. Use the `<TVOConnectOptions>` builder to specify Call parameters and UUID.
 
   The `<TVOCallDelegate>` will receive the Call state update callbacks. The callbacks are same as `connectWithAccessToken:delegate:` API.
 
   If connect fails, `call:didFailToConnectWithError` or `call:didDisconnectWithError` callback is raised with an `NSError` object. `[error localizedDescription]`,
   `[error localizedFailureReason]` and `[error code]` provide details of the failure. The error codes are same as `connectWithAccessToken:delegate:` API.
 
   **Insights**
 
   <table border="1" summary="Insights events.">
   <tr>
   <td>Group Name</td><td>Event Name</td><td>Description</td><td>Since version</td>
   </tr>
   <tr>
   <td>connection</td><td>outgoing</td><td>Outgoing call is made</td><td>3.0.0-beta2</td>
   </tr>
   </table>
 
   If connection fails then an error event will be published.
 
   <table border="1" summary="Insights events.">
   <tr>
   <td>Group Name</td><td>Event Name</td><td>Description</td><td>Since version</td>
   </tr>
   <tr>
   <td>connection</td><td>error</td><td>error description</td><td>3.0.0-beta2</td>
   </tr>
   </table>
 
   ** An example of using the *connectWithAccessToken:delegate* API **

 
   ```
    @interface CallViewController () <TVOCallDelegate, ... >
 
    // property declarations
 
    @end
 
    @implementation CallViewController
 
    // ViewController setup and other code ...
 
    - (void)makeCall:(NSString *)accessToken uuid:(NSString *)uuid region:(NSString *)region to:(NSString *)phoneNumber
        NSMutableDictionary *params = [NSMutableDictionary dictionary];
 
        if ([to length] > 0) {
            params[@"Mode"] = @"Voice";
            params[@"PhoneNumber"] = phoneNumber;
        }
 
        TVOConnectOptions *connectOptions = [TVOConnectOptions optionsWithAccessToken:accessToken
            block:^(TVOConnectOptionsBuilder *builder) {
                builder.params = params;
                builder.uuid = uuid;
                builder.region = region;
        }];
 
        self.call = [TwilioVoice connectWithOptions:connectOptions delegate:self];
    }
 
    #pragma mark - TVOCallDelegate
 
    - (void)callDidStartRinging:(TVOCall *)call {
        NSLog(@"Call is ringing at called party %@", call.to);
    }
 
    - (void)call:(TVOCall *)call didFailToConnectWithError:(NSError *)error {
        NSLog("Failed to Connect the Call: %@\n\t  - reason: %@", [error localizedDescription], [error localizedFailureReason]);
    }
 
    - (void)call:(TVOCall *)call didDisconnectWithError:(NSError *)error {
        if (error) {
            NSLog(@"Call disconnected with error: %@", error);
        } else {
            NSLog(@"Call disconnected");
        }
    }
 
    - (void)callDidConnect:(TVOCall *)call {
        NSLog(@"Call connected.");
    }
 
 @end
 ```
 
   @param options The connect options.
   @param delegate A `<TVOCallDelegate>` to receive Call state updates.
 
   @return A `<TVOCall>` object.
 
   @see TVOCall
   @see TVOCallDelegate
   @see TVOConnectOptions
 */
+ (nonnull TVOCall *)connectWithOptions:(nonnull TVOConnectOptions *)options
                               delegate:(nonnull id<TVOCallDelegate>)delegate;

- (null_unspecified instancetype)init __attribute__((unavailable("TwilioVoice cannot be instantiated directly.")));

@end


/**
 * CallKit Audio Session Handling
 */
@interface TwilioVoice (CallKitIntegration)

@end
